@@ -31,7 +31,6 @@
 #
 
 
-import itertools
 import operator
 from typing import (
     Callable,
@@ -39,13 +38,13 @@ from typing import (
     Tuple,
     Optional,
     Iterable,
-    MutableSequence,
     Collection,
     Sequence,
 )
 
 UnaryPredicate = Callable[[Any], bool]
 """A callable that takes one argument and returns a bool"""
+
 BinaryPredicate = Callable[[Any, Any], bool]
 """A callable that takes two arguments and returns a bool"""
 
@@ -54,84 +53,76 @@ UnaryFunction = Callable[[Any], None]
 
 
 def all_of(iterable: Iterable, unary_predicate: UnaryPredicate) -> bool:
-    """Check if an unary predicate returns True for all elements in the iterable, or if the iterable is empty.
+    """Check if an unary predicate returns True for all elements in the iterable.
 
     Args:
         iterable: An iterable to apply the unary_predicate to
         unary_predicate: An unary predicate to apply to each element in the iterable
 
     Returns:
-        True if the predicate evaluates to True for every element in the iterable, or the iterable is empty, False otherwise
+        True if the predicate evaluates to True for every element in the iterable, False otherwise or if the iterable is empty
     """
-    return all(map(unary_predicate, iterable))
+    return False if not iterable else all(map(unary_predicate, iterable))
 
 
 def any_of(iterable: Iterable, unary_predicate: UnaryPredicate) -> bool:
-    """Check if an unary predicate returns True for any elements in the iterable, or if the iterable is empty.
+    """Check if an unary predicate returns True for any elements in the iterable.
 
     Args:
         iterable: An iterable to apply the unary_predicate to
         unary_predicate: An unary predicate to apply to each element in the iterable
 
     Returns:
-        True if the predicate evaluates to True for any element in the iterable, False otherwise or when the iterable is empty
+        True if the predicate evaluates to True for any element in the iterable, False otherwise or if the iterable is empty
     """
-    return any(map(unary_predicate, iterable))
+    return False if not iterable else any(map(unary_predicate, iterable))
 
 
 def none_of(iterable: Iterable, unary_predicate: UnaryPredicate) -> bool:
-    """Check if an unary predicate returns False for any elements in the iterable, or if the iterable is empty.
+    """Check if an unary predicate returns True for no elements in the iterable.
 
     Args:
         iterable: An iterable to apply the unary_predicate to
         unary_predicate: An unary predicate to apply to each element in the iterable
 
     Returns:
-        True if the predicate evaluates to False for every element in the iterable, False otherwise or when the iterable is empty
+        True if the predicate evaluates to True for no element in the iterable or if the iterable is empty, False otherwise
     """
     return True if not iterable else not all(map(unary_predicate, iterable))
 
 
 def for_each(
-    mutable_sequence: MutableSequence,
+    iterable: Iterable,
     unary_function: UnaryFunction,
-    readonly: bool = True,
 ) -> None:
-    """Apply an unary function to each element in the mutable_sequence, modifying the elements.
+    """Apply an unary function to each element in the iterable.
+
+    This function has read-only access to the iterable
 
     Args:
-        mutable_sequence: A mutable iterable to apply the unary_function to
-        unary_function: An unary function to apply to each element in the mutable_sequence
-        readonly: indicates whether the mutable_sequence is intended to be read or modified,defaults to True
+        iterable: An iterable to apply the unary_function to
+        unary_function: An unary function to apply to each element in the iterable
     """
-    if readonly:
-        for element in mutable_sequence:
-            unary_function(element)
-    else:
-        mutable_sequence[:] = list(map(unary_function, mutable_sequence))
+    for element in iterable:
+        unary_function(element)
 
 
 def for_each_n(
-    mutable_sequence: MutableSequence,
+    sequence: Sequence,
     unary_function: UnaryFunction,
     num_elements: int,  # noqa: VNE001,C0103
-    readonly: bool = True,
 ) -> None:
-    """Apply an unary function to the first num_elements elements in the mutable_sequence, modifying the elements.
+    """Apply an unary function to the first num_elements elements in the sequence, modifying the elements.
+
+    This function has read-only access to the sequence
 
     Args:
-        mutable_sequence: A mutable iterable to apply the unary_function to
-        unary_function: An unary function to apply to each element in the mutable_sequence
+        sequence: A sequence to apply the unary_function to
+        unary_function: An unary function to apply to each element in the sequence
         num_elements: A number indicating the number of elements (counted from the start) to apply the unary_function to
-        readonly: indicates whether the mutable_sequence is intended to be read or modified,defaults to True
     """
-    if readonly:
-        for element in mutable_sequence[:num_elements]:
-            unary_function(element)
-    else:
-        mutable_sequence[:num_elements] = list(
-            map(unary_function, mutable_sequence[:num_elements])
-        )
+    for element in sequence[:num_elements]:
+        unary_function(element)
 
 
 def count(sequence: Sequence, target: Any) -> int:
@@ -139,7 +130,7 @@ def count(sequence: Sequence, target: Any) -> int:
 
     Args:
         sequence: An iterable in which occurrences of target are counted
-        target: A value/object, which's occurrences will be counted
+        target: A value/object, which occurrences will be counted
 
     Returns:
         How often target appeared in sequence
@@ -160,21 +151,17 @@ def count_if(iterable: Iterable, unary_predicate: UnaryPredicate) -> int:
     return sum(map(unary_predicate, iterable))
 
 
-def count_if_not(iterable: Iterable, unary_predicate: UnaryPredicate) -> int:
-    """Count for how many elements in a iterable an unary predicate returns False.
+def count_if_not(collection: Collection, unary_predicate: UnaryPredicate) -> int:
+    """Count for how many elements in a collection an unary predicate returns False.
 
     Args:
-        iterable: An iterable for which to count for how many elements unary_predicate returns False
-        unary_predicate: A value/object, for which to count for how many items in the iterable it returns False
+        collection: An collection for which to count for how many elements unary_predicate returns False
+        unary_predicate: A value/object, for which to count for how many items in the collection it returns False
 
     Returns:
-        For how many items unary predicate returned False, -1 if the iterable is empty
+        For how many items unary predicate returned False
     """
-    return (
-        -1
-        if not iterable
-        else sum(map(lambda element: not unary_predicate(element), iterable))
-    )
+    return len(collection) - sum(map(unary_predicate, collection))
 
 
 def mismatch(
@@ -190,7 +177,7 @@ def mismatch(
         sequence1: First sequence to use for comparison
         sequence2: Second sequence to use for comparison
         binary_predicate: A binary predicate, that returns true if the elements from both sequences are considered equal
-            If not provided, operator.eq will be used as a binary_predicate
+           Defaults to: operator.eq
 
     Returns:
         None, if one or more sequences is empty, or they do not differ until the end of the shortest sequence.
@@ -203,72 +190,78 @@ def mismatch(
 
         mismatch([1, 2, 3, 4, 5], [1, 2, 3, 4, 5]) returns None, since both sequences are considered equal
 
-        mismatch([1, 2, 3, 4, 5], [1, 2, 3, 4, 6]) returns (5, 6), since  their fifth index is considered not equal
+        mismatch([1, 2, 3, 4, 5], [1, 2, 3, 4, 6]) returns (5, 6), the values at the fifth index, since  their fifth index is considered not equal
 
-        mismatch([1, 2, 3, 4, 5], [1, 2, 3, 4, 6, 7, 8, 9]) returns (5, 6), since  their fifth index is considered not equal.
-        Since one sequence is longer, it's additional elements are not compared
+        mismatch([1, 2, 3, 4, 5], [1, 2, 3, 4, 6, 7, 8, 9]) returns (5, 6) for the same reason.
+            Since one sequence is longer, it's additional elements are not compared
     """
     if not sequence1 or not sequence2:
         return None
-    last_index_shortest_sequence: int = min(len(sequence1), len(sequence2)) - 1
 
     return next(
         (pair for pair in zip(sequence1, sequence2) if not binary_predicate(*pair)),
-        (
-            sequence1[last_index_shortest_sequence],
-            sequence2[last_index_shortest_sequence],
-        ),
+        None,
     )
 
 
-def find(collection: Collection, target_element: Any) -> int:
-    """Find the index of the first occurrence of target_element in collection.
+def find(sequence: Sequence, target_element: Any) -> int:
+    """Find the index of the first occurrence of target_element in sequence.
 
     Args:
-        collection: A collection which to search through
-        target_element: An element to search in the collection
+        sequence: A sequence which to search through
+        target_element: An element to search in the sequence
 
     Returns:
-        The index of target_element's first occurrence, -1 if it was not found
+        The index of target_element's first occurrence, -1 if it was not found or the sequence is empty
     """
+    if not sequence:
+        return -1
+
     try:
-        return list(
-            map(operator.eq, collection, itertools.repeat(target_element))
-        ).index(True)
+        return sequence.index(target_element)
+
     except ValueError:
-        return len(collection) - 1
+        return -1
 
 
-def find_if(collection: Collection, unary_predicate: UnaryPredicate) -> int:
-    """Find the index of the first element in collection satisfying unary_predicate.
+def find_if(iterable: Iterable, unary_predicate: UnaryPredicate) -> int:
+    """Find the index of the first element in iterable satisfying unary_predicate.
 
     Args:
-        collection: A collection which to search through
+        iterable: A iterable which to search through
         unary_predicate: An UnaryPredicate, which determines if the current value is our target
 
     Returns:
-        The index of the first element which satisfies unary_predicate, -1 if no element satisfies unary_predicate
+        The index of the first element which satisfies unary_predicate, -1 if no element satisfies unary_predicate or the iterable is empty
     """
+    if not iterable:
+        return -1
+
     try:
-        return list(map(unary_predicate, collection)).index(True)
+        return list(map(unary_predicate, iterable)).index(True)
+
     except ValueError:
-        return len(collection) - 1
+        return -1
 
 
-def find_if_not(collection: Collection, unary_predicate: UnaryPredicate) -> int:
-    """Find the index of the first element in collection NOT satisfying unary_predicate.
+def find_if_not(iterable: Iterable, unary_predicate: UnaryPredicate) -> int:
+    """Find the index of the first element in iterable NOT satisfying unary_predicate.
 
     Args:
-        collection: A collection which to search through
+        iterable: A iterable which to search through
         unary_predicate: An UnaryPredicate, which determines if the current value is NOT our target
 
     Returns:
-        The index of the first element which DOES NOT satisfy unary_predicate, -1 if all elements satisfy unary_predicate
+        The index of the first element which DOES NOT satisfy unary_predicate, -1 if all elements satisfy unary_predicate or the iterable is empty
     """
+    if not iterable:
+        return -1
+
     try:
-        return list(map(unary_predicate, collection)).index(False)
+        return list(map(unary_predicate, iterable)).index(False)
+
     except ValueError:
-        return len(collection) - 1
+        return -1
 
 
 def find_end(
@@ -281,13 +274,16 @@ def find_end(
     Args:
         collection_super: A Collection in which to search for the collection_sub
         collection_sub: A Collection to search for in collection_super
-        binary_predicate: A BinaryPredicate used to check if an index's elements of both collections are considered equal
+        binary_predicate: A BinaryPredicate used to check if an index's elements of both collections are considered equal, defaults to: operator.eq
 
     Returns:
         The index of the beginning of the last occurrence of collection_sub in collection_super,
-            or the last index in collection_super if it is empty or collection_sub is not found in it
+            -1 if any of the two collections is empty, or collection_sub does not occur once in collection_super
 
     """
+    if not collection_super or not collection_sub:
+        return -1
+
     for i in range(  # noqa: VNE001
         (len(collection_super) - 1) - (len(collection_sub) - 1), -1, -1
     ):
@@ -296,32 +292,34 @@ def find_end(
         ):
             return i
 
-    return len(collection_super) - 1
+    return -1
 
 
 def find_first_of(
-    values_in: Collection,
-    values_from: Collection,
+    iterable_super: Iterable,
+    iterable_sub: Iterable,
     binary_predicate: BinaryPredicate = operator.eq,
 ) -> int:
-    """Find first index in values_in at which an element of values_from occurs.
+    """Find first index in values_in at which an element of iterable_sub occurs.
 
     Args:
-        values_in: A Collection in which to search for values of values_from
-        values_from: A Collection of values to search for in values_in
-        binary_predicate: A BinaryPredicate used to check if an index's elements of both collections are considered equal
+        iterable_super: A Iterable in which to search for values of iterable_sub
+        iterable_sub: A Iterable of values to search for in values_in
+        binary_predicate: A BinaryPredicate used to check if an index's elements of both Iterables are considered equal, defaults to: operator.eq
 
     Returns:
-        The first index in values_in at which an element of values_from occurs
-            or the last index in  values_in if it is empty or no element from values_from is found in it
-
+        The first index in values_in at which an element of iterable_sub occurs,
+            or -1 if any of the two iterables is empty or iterable_sub does not occur once in iterable_super
     """
-    for i, element_in in enumerate(values_in):  # noqa: VNE001
-        for element_from in values_from:
-            if binary_predicate(element_in, element_from):
+    if not iterable_super or not iterable_sub:
+        return -1
+
+    for i, element_super in enumerate(iterable_super):  # noqa: VNE001
+        for element_sub in iterable_sub:
+            if binary_predicate(element_super, element_sub):
                 return i
 
-    return len(values_in) - 1
+    return -1
 
 
 def adjacent_find(
@@ -331,13 +329,11 @@ def adjacent_find(
 
     Args:
         sequence: a sequence to search through
-        binary_predicate: a binary predicate to evaluate the equality of adjacent elements
+        binary_predicate: a binary predicate to evaluate the equality of adjacent elements, defaults to: operator.eq
 
     Returns:
         The first index at which two adjacent elements are considered equal,
-            or the last index, if no two adjacent elements are considered equal,
-            or -1 if the sequence is empty
-
+            or -1 if the sequence is empty or no two adjacent elements are considered equal
     """
     if not sequence:
         return -1
@@ -348,7 +344,7 @@ def adjacent_find(
             list(map(binary_predicate, sequence[::2], sequence[1::2])).index(True) * 2
         )
     except ValueError:
-        return len(sequence) - 1
+        return -1
 
 
 def search(
@@ -361,14 +357,13 @@ def search(
     Args:
         sequence_super: A sequence to search in
         sequence_sub: A sequence to search for in sequence_super
-        binary_predicate: A binary predicate to evaluate the equality of compared items
+        binary_predicate: A binary predicate to evaluate the equality of compared items, defaults to: operator.eq
 
     Returns:
-        The index of the beginning of the first occurrence of sequence_sub in sequence_super
-            If not found, returns the last index of sequence_super
-            If any sequence is empty, returns -1
+        The index of the beginning of the first occurrence of sequence_sub in sequence_super,
+            or -1 if any sequence_super or sequence_sub is empty or sequence_sub does not occur once in sequence_super
     """
-    if not sequence_sub:
+    if not sequence_super or not sequence_sub:
         return -1
 
     for i in range(len((sequence_super))):  # noqa: VNE001
@@ -378,7 +373,7 @@ def search(
 
             return i
 
-    return len(sequence_super) - 1
+    return -1
 
 
 def search_n(
@@ -393,12 +388,11 @@ def search_n(
         sequence: A sequence to search in
         value: any object to search for in collection
         num_elements: How many times, counting from the beginning, value has to be repeated
-        binary_predicate: A binary predicate to evaluate the equality of consecutive items
+        binary_predicate: A binary predicate to evaluate the equality of consecutive items, defaults to: operator.eq
 
     Returns:
-        The index of the beginning of the first num_elements repetitions of value in sequence
-            If not found, returns the last index of sequence_super
-            If sequence is empty, returns -1
+        The index of the beginning of the first num_elements repetitions of value in sequence,
+            or -1 if sequence is empty or value does not occur once in sequence
     """
     if not sequence:
         return -1
@@ -410,4 +404,50 @@ def search_n(
         else:
             return i
 
-    return len(sequence) - 1
+    return -1
+
+
+def copy_replace(iterable: Iterable, old_val: Any, new_val: Any) -> Iterable:
+    """Copy iterable while replacing all occurrences of old_val with new_val.
+
+    Args:
+        iterable: An iterable to copy
+        old_val: A value to replace
+        new_val: A value serving as the replacement
+
+    Returns:
+        A generator yielding the values of iterable with all occurrences of old_val replaced with new_val
+    """
+    return (val if val != old_val else new_val for val in iterable)
+
+
+def copy_replace_if(
+    iterable: Iterable, unary_predicate: UnaryPredicate, new_val: Any
+) -> Iterable:
+    """Copy iterable while replacing all values satisfying unary_predicate with new_val.
+
+    Args:
+        iterable: An iterable to copy
+        unary_predicate: An unary predicate deciding whether to replace an item
+        new_val: A value serving as the replacement
+
+    Returns:
+        A generator yielding the values of iterable with all values satisfying unary_predicate replaced with new_val
+    """
+    return (new_val if unary_predicate(val) else val for val in iterable)
+
+
+def copy_replace_if_not(
+    iterable: Iterable, unary_predicate: UnaryPredicate, new_val: Any
+) -> Iterable:
+    """Copy iterable while replacing all values not satisfying unary_predicate with new_val.
+
+    Args:
+        iterable: An iterable to copy
+        unary_predicate: An unary predicate deciding whether to replace an item
+        new_val: A value serving as the replacement
+
+    Returns:
+        A generator yielding the values of iterable with all values not satisfying unary_predicate replaced with new_val
+    """
+    return (new_val if not unary_predicate(val) else val for val in iterable)
